# SPDX-FileCopyrightText: 2023-present Dynatrace LLC
#
# SPDX-License-Identifier: MIT

from collections.abc import Callable
from datetime import datetime, timedelta

from .activation import ActivationConfig, ActivationType
from .communication import Status
from .event import Severity
from .extension import DtEventType, Extension
from .metric import MetricType, SummaryStat


class _HelperExtension(Extension):
    @property
    def is_helper(self) -> bool:
        return True


def report_metric(
    key: str,
    value: float | str | int | SummaryStat,
    dimensions: dict[str, str] | None = None,
    techrule: str | None = None,
    timestamp: datetime | None = None,
    metric_type: MetricType = MetricType.GAUGE,
) -> None:
    """Reports a metric using the MINT protocol
    By default, it reports a gauge metric


    :param key: The metric key, must follow the MINT specification
    :param value: The metric value, can be a simple value or a SummaryStat
    :param dimensions: A dictionary of dimensions
    :param techrule: The techrule of the metric, defaults to None
    :param timestamp: The timestamp of the metric, defaults to the current time
    :param metric_type: The type of the metric, defaults to MetricType.GAUGE
    """
    _HelperExtension().report_metric(key, value, dimensions, techrule, timestamp, metric_type)  # type: ignore


def report_mint_lines(lines: list[str]) -> None:
    """Reports mint lines using the MINT protocol.
    These lines are not validated before being sent.

    :param lines: A list of mint lines, example: ["my_metric 1", "my_other_metric 2"]
    """
    _HelperExtension().report_mint_lines(lines)


def report_dt_event(
    event_type: DtEventType,
    title: str,
    start_time: int | None = None,
    end_time: int | None = None,
    timeout: int | None = None,
    entity_selector: str | None = None,
    properties: dict[str, str] | None = None,
) -> None:
    """
    Reports a custom event v2 using event ingest

    For reference see: https://www.dynatrace.com/support/help/dynatrace-api/environment-api/events-v2/post-event

    :param event_type: The event type chosen from type Enum (required)
    :param title: The title of the event (required)
    :param start_time: The start time of event in UTC ms, if not set, current timestamp (optional)
    :param end_time: The end time of event in UTC ms, if not set, current timestamp + timeout (optional)
    :param timeout: The timeout of event in minutes, if not set, 15 (optional)
    :param entity_selector: The entity selector, if not set, the event is associated with environment entity (optional)
    :param properties: A map of event properties (optional)
    """
    _HelperExtension().report_dt_event(event_type, title, start_time, end_time, timeout, entity_selector, properties)


def report_dt_event_dict(event: dict):
    """
    Reports a custom event v2 using event ingest using provided dictionary resembling required json

    For reference see: https://www.dynatrace.com/support/help/dynatrace-api/environment-api/events-v2/post-event
    """
    _HelperExtension().report_dt_event_dict(event)


def schedule(
    callback: Callable,
    interval: timedelta | int,
    args: tuple | None = None,
    activation_type: ActivationType | None = None,
) -> None:
    """Schedules a callback to be called periodically

    :param callback: The callback to be called
    :param interval: The time interval between invocations, can be a timedelta object, or an int representing the number of seconds
    :param args: Arguments to the callback, if any
    :param activation_type: Optional activation type when this callback should run, can be 'ActivationType.LOCAL' or 'ActivationType.REMOTE'

    """
    _HelperExtension().schedule(callback, interval, args, activation_type)


def schedule_function(
    interval: timedelta | int, args: tuple | None = None, activation_type: ActivationType | None = None
):
    def decorator(function):
        schedule(function, interval, args=args, activation_type=activation_type)

    return decorator


def schedule_method(
    interval: timedelta | int, args: tuple | None = None, activation_type: ActivationType | None = None
):
    def decorator(function):
        Extension.schedule_decorators.append((function, interval, args, activation_type))

    return decorator


def report_event(
    title: str,
    description: str,
    properties: dict | None = None,
    timestamp: datetime | None = None,
    severity: Severity | str = Severity.INFO,
) -> None:
    """Reports an event using the MINT protocol

    :param title: The title of the event
    :param description: The description of the event
    :param properties: A dictionary of properties
    :param timestamp: The timestamp of the event, defaults to the current time
    :param severity: The severity of the event, defaults to Severity.INFO
    """
    _HelperExtension().report_event(title, description, properties, timestamp, severity)


def report_log_event(log_event: dict):
    """Reports a custom log event using log ingest

    :param log_event: The log event dictionary, reference: https://www.dynatrace.com/support/help/shortlink/log-monitoring-log-data-ingestion
    """
    _HelperExtension().report_log_event(log_event)


def report_log_events(log_events: list[dict]):
    """Reports a list of custom log events using log ingest

    :param log_events: The list of log events
    """
    _HelperExtension().report_log_events(log_events)


def report_log_lines(log_lines: list[str | bytes]):
    """Reports a list of log lines using log ingest

    :param log_lines: The list of log lines
    """
    _HelperExtension().report_log_lines(log_lines)


def run_extension():
    """Starts the extension loop"""
    _HelperExtension().run()


def get_activation_config():
    return _HelperExtension().get_activation_config()


def get_helper_extension():
    return _HelperExtension()


def dt_fastcheck():
    def wrapper(fast_check_callback: Callable[[ActivationConfig, str], Status]):
        _HelperExtension().register_fastcheck(fast_check_callback=fast_check_callback)

    return wrapper


def log(message: str, severity: Severity):
    """
    Logs provided message into the extension's log file.
    Only two severities are supported: INFO and ERROR
    """
    if severity == Severity.INFO:
        _HelperExtension().logger.info(message)
    elif severity == Severity.ERROR:
        _HelperExtension().logger.error(message)



# SPDX-FileCopyrightText: 2023-present Dynatrace LLC
#
# SPDX-License-Identifier: MIT

from enum import Enum


class ActivationType(Enum):
    REMOTE = "remote"
    LOCAL = "local"


class ActivationConfig(dict):
    def __init__(self, activation_context_json: dict):
        self._activation_context_json = activation_context_json
        self.version: str = self._activation_context_json.get("version", "")
        self.enabled: bool = self._activation_context_json.get("enabled", True)
        self.description: str = self._activation_context_json.get("description", "")
        self.feature_sets: list[str] = self._activation_context_json.get("featureSets", [])
        self.type: ActivationType = ActivationType.REMOTE if self.remote else ActivationType.LOCAL
        super().__init__()

    @property
    def config(self) -> dict:
        return self.remote if self.remote else self.local

    @property
    def remote(self) -> dict:
        return self._activation_context_json.get("pythonRemote", {})

    @property
    def local(self) -> dict:
        return self._activation_context_json.get("pythonLocal", {})

    def __getitem__(self, item):
        return self.config[item]

    def get(self, key, default=None):
        return self.config.get(key, default)

    def __repr__(self):
        return f"ActivationConfig(version='{self.version}', enabled={self.enabled}, description='{self.description}', type={self.type}, config={self.config})"



import json
import unittest
from unittest.mock import MagicMock, mock_open, patch

from dynatrace_extension.sdk.communication import (
    MAX_LOG_REQUEST_SIZE,
    MAX_METRIC_REQUEST_SIZE,
    HttpClient,
    divide_into_batches,
)


class TestCommunication(unittest.TestCase):
    @patch("builtins.open", mock_open(read_data="test_token"))
    @patch.object(HttpClient, "_make_request", return_value=MagicMock())
    def test_http_client_metric_report(self, mock_make_request):
        http_client = HttpClient("https://localhost:9999", "1", "token", MagicMock())
        few_metrics = ["metric 1", "metric 2"]
        responses = http_client.send_metrics(few_metrics)
        self.assertEqual(len(responses), 1)

        many_metrics = ['my.metric,dim="dim" 10'] * 500 * 100
        responses = http_client.send_metrics(many_metrics)
        self.assertEqual(len(responses), 2)

        no_metrics = []
        responses = http_client.send_metrics(no_metrics)
        self.assertEqual(len(responses), 0)

    def test_large_log_chunk(self):
        # This is 14_660_000 bytes
        events = []
        for _ in range(5000):
            attributes = {}
            for j in range(150):
                attributes[f"attribute{j}"] = j
            events.append(attributes)

        # it needs to be divided into 4 lists, each with 3_665_000 bytes
        chunks = list(divide_into_batches(events, MAX_LOG_REQUEST_SIZE))
        self.assertEqual(len(chunks), 4)
        self.assertEqual(len(chunks[0]), 3665000)
        self.assertEqual(len(chunks[1]), 3665000)
        self.assertEqual(len(chunks[2]), 3665000)
        self.assertEqual(len(chunks[3]), 3665000)

    def test_small_log_chunk(self):
        events = []
        for _ in range(10):
            attributes = {}
            for j in range(10):
                attributes[f"attribute{j}"] = j
            events.append(attributes)

        chunks = list(divide_into_batches(events, MAX_LOG_REQUEST_SIZE))
        self.assertEqual(len(chunks), 1)
        self.assertEqual(len(chunks[0]), 1720)

    def test_large_metric_chunk(self):
        metrics = ['my.metric,dim="dim" 10'] * 500 * 100

        # it needs to be divided into 2 lists, each with 650_000 bytes
        chunks = list(divide_into_batches(metrics, MAX_METRIC_REQUEST_SIZE, "\n"))
        self.assertEqual(len(chunks), 2)
        self.assertEqual(len(chunks[0]), 574999)
        self.assertEqual(len(chunks[1]), 574999)

    def test_small_metric_chunk(self):
        metrics = ['my.metric,dim="dim" 10'] * 100

        chunks = list(divide_into_batches(metrics, MAX_METRIC_REQUEST_SIZE, "\n"))
        self.assertEqual(len(chunks), 1)
        self.assertEqual(len(chunks[0]), 2299)

    def test_no_metrics(self):
        metrics = []

        chunks = list(divide_into_batches(metrics, MAX_METRIC_REQUEST_SIZE, "\n"))
        self.assertEqual(len(chunks), 0)

    def test_large_log_chunk_valid_json(self):
        events = []
        for _ in range(5000):
            attributes = {}
            for j in range(150):
                attributes[f"attribute{j}"] = j
            events.append(attributes)

        # it needs to be divided into 4 lists, each with 3_665_000 bytes
        chunks = list(divide_into_batches(events, MAX_LOG_REQUEST_SIZE))
        self.assertEqual(len(chunks), 4)

        for chunk in chunks:
            json.loads(chunk)



# SPDX-FileCopyrightText: 2023-present Dynatrace LLC
#
# SPDX-License-Identifier: MIT

import logging
from typing import ClassVar, NamedTuple


class DefaultLogLevel(NamedTuple):
    string_value: str
    int_value: int


class RuntimeProperties:
    _default_log_level = DefaultLogLevel("info", logging.INFO)
    _log_level_converter: ClassVar = {"debug": logging.DEBUG, "info": logging.INFO}

    def __init__(self, json_response: dict):
        """
        This is the response from EEC when a status (heartbeat) is sent
        Example:
        {'extconfig': 'b2520a74-88e8-3e03-bc01-e1116fec4a98', 'userconfig': '1645918226657', 'debugmode': '0', 'runtime': {}, 'tasks': []}
        """
        self.extconfig: str = json_response.get("extconfig", "")
        self.userconfig: str = json_response.get("userconfig", "")
        self.debugmode: bool = json_response.get("debugmode", "0") == "1"
        self.runtime: dict = json_response.get("runtime", {})
        self.tasks: list[str] = json_response.get("tasks", [])

    @classmethod
    def set_default_log_level(cls, value: str):
        RuntimeProperties._default_log_level = DefaultLogLevel(value, RuntimeProperties._to_log_level(value))

    @classmethod
    def _to_log_level(cls, value: str) -> int:
        """
        The method convert LogLevel string value into Python log level (loggin package).
        loggin.INFO is a default.
        :param value: string log lever
        :return: Python log level
        """
        return RuntimeProperties._log_level_converter.get(value, RuntimeProperties._default_log_level.int_value)

    def log_level(self, extension_name: str) -> int:
        """
        The method check python.debuglevel (lower priority)
        and python.debuglevel.extension_name (higher priority) string debug flags.
        loggin.INFO is a default.
        :param extension_name: extension name
        :return: log level for Python log system (loggin)
        """
        value = self.runtime.get("debuglevel", RuntimeProperties._default_log_level.string_value)
        value = self.runtime.get(f"debuglevel.{extension_name}", value)
        return RuntimeProperties._to_log_level(value)

    def get_api_log_level(self, extension_name: str) -> int:
        """
        The method check python.debuglevel.api (lower priority)
        python.debuglevel.extension_name.api (higher priority) string debug flags.
        loggin.INFO is a default.
        :param extension_name: extension name
        :return: log level for Python log system (loggin)
        """
        value = self.runtime.get("debuglevel.api", RuntimeProperties._default_log_level.string_value)
        value = self.runtime.get(f"debuglevel.{extension_name}.api", value)
        return RuntimeProperties._to_log_level(value)
        pass
